import { computed, reactive, ref, unref } from 'vue';
import { LinkedList, tryGet, trySet } from '../common';
import { FormField } from './FormField';
const isSimpleRule = (rule) => typeof rule === 'function';
export class Form {
    constructor() {
        this.simpleMap = new Map();
        this.keyedSetMap = new Map();
        this.reactiveFormFieldMap = reactive(new Map());
        this.trySetKeyedSet = trySet(this.keyedSetMap);
        this.tryGetKeyedSet = tryGet(this.keyedSetMap);
        this.tryGetSimple = tryGet(this.simpleMap);
        this.submitting = ref(false);
    }
    registerField(uid, name, modelValue, rules) {
        const formField = new FormField(name, modelValue, rules);
        const simple = {
            formField,
            keys: [],
            vs: [],
            rollbacks: []
        };
        rules.forEach((rule, ruleNumber) => {
            const validate = Form.validateFactory(formField, rule, ruleNumber);
            if (!validate) {
                return;
            }
            if (isSimpleRule(rule)) {
                simple.vs.push(validate);
            }
            else {
                const { key } = rule;
                const keyed = {
                    formField,
                    v: validate
                };
                const rollback = () => {
                    this.tryGetKeyedSet({
                        success: keyedSet => {
                            keyedSet.delete(keyed);
                            if (keyedSet.size === 0) {
                                this.keyedSetMap.delete(key);
                            }
                        }
                    })(key);
                };
                simple.keys.push(key);
                simple.rollbacks.push(rollback);
                this.trySetKeyedSet({
                    failure: keyedSet => keyedSet.add(keyed)
                })(key, new Set([keyed]));
            }
        });
        this.simpleMap.set(uid, simple);
        this.reactiveFormFieldMap.set(uid, formField);
        return formField;
    }
    validate(uid) {
        const simple = this.simpleMap.get(uid);
        if (simple && simple.formField.touched) {
            return Promise.allSettled([
                ...simple.vs.map(v => v([simple.formField.modelValue])),
                ...this.getPromisesFor(simple.keys)
            ]);
        }
    }
    async validateAll() {
        const promises = [];
        for (const { formField, vs } of this.simpleMap.values()) {
            formField.touched = true;
            promises.push(...vs.map(v => v([formField.modelValue])));
        }
        promises.push(...this.getPromisesFor(this.keyedSetMap.keys()));
        const settledResults = await Promise.allSettled(promises);
        for (const result of settledResults) {
            if (result.status === 'rejected') {
                return true;
            }
        }
        return false;
    }
    onDelete(uid) {
        this.tryGetSimple({
            success({ rollbacks }) {
                rollbacks.forEach(r => r());
            }
        })(uid);
        this.simpleMap.delete(uid);
        this.reactiveFormFieldMap.delete(uid);
    }
    getErrors() {
        return computed(() => {
            const errors = [];
            for (const formField of this.reactiveFormFieldMap.values()) {
                errors.push(...formField.getErrors().value);
            }
            return errors;
        });
    }
    resetFields(toDefaultValues = true) {
        for (const { formField } of this.simpleMap.values()) {
            formField.reset(toDefaultValues);
        }
    }
    getPromisesFor(keys) {
        const promises = [];
        for (const key of keys) {
            this.tryGetKeyedSet({
                success: keyedSet => {
                    if (this.isEveryFormFieldTouchedWith(key)) {
                        const values = [...keyedSet.values()];
                        const modelValues = values.map(({ formField }) => formField.modelValue);
                        const vs = values.map(({ v }) => v(modelValues));
                        promises.push(...vs);
                    }
                }
            })(key);
        }
        return promises;
    }
    isEveryFormFieldTouchedWith(key) {
        let everyFormFieldIsTouched = true;
        this.tryGetKeyedSet({
            success: keyedSet => {
                for (const { formField } of keyedSet) {
                    if (!formField.touched) {
                        everyFormFieldIsTouched = false;
                        break;
                    }
                }
            }
        })(key);
        return everyFormFieldIsTouched;
    }
    static validateFactory(formField, rule, ruleNumber) {
        const buffer = new LinkedList();
        const setError = (formField, ruleNumber, error) => {
            if (typeof error === 'string' && formField.touched) {
                formField.setError(ruleNumber, error);
                throw error;
            }
            else {
                formField.setError(ruleNumber, null);
            }
        };
        const validator = (formField, rule, ruleNumber) => async (modelValues) => {
            let error;
            const ruleResult = rule(...modelValues.map(unref));
            if (typeof (ruleResult === null || ruleResult === void 0 ? void 0 : ruleResult.then) === 'function') {
                formField.validating.value = true;
                const node = buffer.addLast(false);
                if (node.prev) {
                    node.prev.value = true;
                }
                try {
                    error = await ruleResult;
                }
                catch (err) {
                    error = err;
                }
                buffer.remove(node);
                if (!node.value) {
                    formField.validating.value = false;
                    setError(formField, ruleNumber, error);
                }
            }
            else {
                error = ruleResult;
                setError(formField, ruleNumber, error);
            }
        };
        if (isSimpleRule(rule)) {
            return validator(formField, rule, ruleNumber);
        }
        else if (rule.rule) {
            return validator(formField, rule.rule, ruleNumber);
        }
    }
}
